/// Import this library in order to get the abstract classes and annotations for
/// the library. Useful to use another implementation without depending on any
/// implementation code within.
library cork.interface;

/// Annotation for denoting a class that can be injected within a [Component].
class Inject {
  const Inject();
}

/// Annotation for denoting a class that includes [Inject]able classes and other
/// defined [Module] annotated classes. Example usage:
///     @Module(const [
///       CoffeeBeanStorage // A class annotated with @Inject.
///       CoffeePumpModule, // A class annotated with @Module.
///       Heater            // A third party class which uses @Provide below.
///     ])
///     class CoffeeMakerModule {
///       @Provide(Heater)
///       Heater getHeater() => return new ThirdPartyHeater();
///     }
class Module {
  /// A collection of injectable [Type]s annotated with `@Inject` (or with a
  /// @Provide method in this module or in an imported module) or `@Module`.
  final List<Type> included;

  const Module(this.included);
}

/// An alternative for 3rd party classes that cannot be annotated with [Inject].
/// A [Module] can define custom provider functions, or have one automatically
/// generated. Example usage:
///
///   // Uses an user-defined factory.
///   @Module(const [Heater])
///   class CoffeeMakerModule {
///     @Provide(Heater)
///     Heater getHeater() => return new ThirdPartyHeater();
///   }
///
///   // Uses an automatically generated factory.
///   @Module(const [Heater])
///   @Provide(Heater)
///   class CoffeeMakerModule {}
class Provide {
  final Type type;

  const Provide(this.type);
}

/// A type of [Module] that should have an [ObjectGraph] created for it in order
/// to create an injector. Example usage:
///
///   @Component(const [CoffeeMakerModule])
///   class CoffeeApplication() {
///     // This will be automatically instantiated by the injector.
///     CoffeeApplication(CoffeeMaker coffeeMaker) {}
///   }
///
/// Unlike module, [Component] also registers itself as a (root) injectable.
class Component extends Module implements Inject {
  const Component(List<Type> included) : super(included);
}

/// A runtime tool that initializes DI bindings.
abstract class ObjectGraph {
  /// Initializes [component] for dependency injection. This can be done using
  /// previous static analysis or dynamically at runtime using reflection.
  create(Type component);
}
