library cork.lib.src.mirrors.utils;

import 'dart:collection';
import 'dart:mirrors';

import 'package:cork/interface.dart';
import 'package:cork/src/injector.dart';

/// Wrapper around [ClassMirror] that exposes a simple factory function that is
/// easily consumable by an injector implementation.
class DynamicTypeFactory<T> implements TypeFactory<T> {
  final ClassMirror _classMirror;
  final Symbol _constructorName;
  List<Type> _parameterTypes;

  DynamicTypeFactory(ClassMirror classMirror) :
      _classMirror = classMirror,
      // TODO: Make this configurable instead of relying on default constructor.
      _constructorName = classMirror.simpleName;

  @override
  T call([List positionalArguments = const []]) {
    return _classMirror.newInstance(
        const Symbol(''),
        positionalArguments)
            .reflectee;
  }

  @override
  Iterable<Type> get parameterTypes {
    if (_parameterTypes == null) {
      MethodMirror methodMirror = _classMirror.declarations[_constructorName];
      assert(methodMirror.isConstructor);
      final types = methodMirror.parameters.map((parameter) {
        return parameter.type.reflectedType;
      }).toList(growable: false);
      _parameterTypes = new UnmodifiableListView<Type>(types);
    }
    return _parameterTypes;
  }
}

/// Helper method that uses mirrors to return a mapping of factory functions.
/// If called with [factories] provided, appends to it, otherwise, a new one is
/// created. The resulting map in either case is also returned.
Map<Type, TypeFactory> extract(
    Type clazz,
    [Map<Type, TypeFactory> factories]) {
  // Read either the @Module/@Component or @Inject annotation.
  var annotation = reflectClass(clazz).metadata.firstWhere((m) {
    return m.reflectee is Module || m.reflectee is Inject;
  }, orElse: () {
    throw new ArgumentError('Valid annotation not found on: $clazz.');
  }).reflectee;

  // If this is the initial call, annotation should be @Component
  if (factories == null) {
    assert(annotation is Component);
    factories = new HashMap<Type, TypeFactory>();
  }

  // Traverse modules
  if (annotation is Module) {
    for (final dependency in annotation.included) {
      extract(dependency, factories);
    }
  }

  // Add directly to the map if an injectable
  if (annotation is Inject) {
    factories[clazz] = new DynamicTypeFactory(reflectClass(clazz));
  }

  return factories;
}
